//
//	Battery discharge capacities meter 0.1
//		main.c
//		2009.04.23 7segment LED support
//		2009.04.24 Battery voltage get
//		2009.04.25 1st rel
//			(c)2009 T.Togashi
//
//	Description:
//		Rechargeble battery capacity measure.
//		Constant resistance discharge method.
//		Display [Wh] but bery simplicity.
//	Device: PIC16F88[LF]
//	Pin assign:
//		p1: RA2 -> LED Cathod-common 1(Left side, sink)
//		p2: RA3 -> LED Cathod-common 2
//		p3: RA4 -> LED Cathod-common 3
//		p4: RA5 <- NC
//		p5: Vss
//		p6: RB0 -> LED Segment D
//		p7: RB1 -> LED Segment dp
//		p8: RB2 -> LED Segment C
//		p9: RB3 -> LED Segment G
//		p10:RB4 -> LED Segment F
//		p11:RB5 -> LED Segment B
//		p12:RB6 -> LED Segment A
//		p13:RB7 -> LED Segment E
//		p14:Vdd
//		p15:RA6 -> FET gate
//		p16:RA7 <- NC
//		p17:AN0 <- High-side Rsense=1[ohm] high point 
//		p18:AN1 <- High-side Rsense=1[ohm] low point
//

#include "htc.h"
#include "adc.h"						// Private library
#include "led7seg.h"					// Private library

#define	bitset(var,bitno)	((var) |= 1UL << (bitno))
#define bitclr(var,bitno)	((var) &= ~(1UL << (bitno)))
#define bittest(var,bitno)	((var) & (1UL << (bitno)))
#define _XTAL_FREQ	4000000
#define VBATLOLIM	100					// 1.00V
#define TIMELIM		345600				// [sec] 4day
#define VREF		3300				// [mV]
#define PLPF		5					// Pusudo LPF rate

//
// Intrrupt handler
//
static unsigned char upd = 0;
static unsigned char tic = 0;
void interrupt isr(void) {
	if (TMR0IE && TMR0IF) {
		TMR0IF=0;
		LED7SEG_Refresh();
		if (tic++ >= (_XTAL_FREQ / 4 / 32 / 256)) {
			tic = 0;
			upd = 1;
		}
	}
}

//
//  M a i n
//
main(){
	long sec;
	long isum;
	long psum;
	int endflg;

//		*** System clock ***
//
//				  OSCCON - Oscillator control register
//				  -0000000  Initial
//				  .XXX....  IRCF: Internal oscillator
//				  .000....     31.25KHz
//				  .001....     125KHz
//				  .010....     250KHz
//				  .011....     500KHz
//				  .100....     1MHz
//				  .101....     2MHz
//				  .110....     4MHz
//				  .111....     8MHz
//				  ....X...  OSTS: Oscillator Start-up Time-out Status bit
//				  ....0...     Device is running from T1OSC or INTRC as secondary system clock
//				  ....1...     Device is running from the primery system clock
//				  .....X..  IOFS:INTOSC Frequency stable bit
//				  .....1..     stable
//				  .....0..     not stable
//				  ......XX  SCS: Oscillator Mode Select bits
//				  ......00    Oscillator mode defined by FOSC<2:0>
//				  ......01    T1OSC is used by system clock
//				  ......10    Internal RC is used for system clock
//				  ......11    Reserved
	OSCCON =    0b01100000;		// 4MHz

//				  OSCTUNE - Oscillator tuning register
//				  00000000  Initial
//				  ..XXXXXX  TUN: Frequency tuning bits
//				  ..011111     +31	+12.5%
//				  ..000001     +1	+0.4%
//				  ..000000     0
//				  ..111111     -1	-0.4%
//				  ..100000     -32	-12.5%
	OSCTUNE = 0;

//		*** System option ***
//
//				  OPTION_REG - Option register
//				  11111111  Initial
//				  X.......  RBPU: PORTB Pull-up enable
//				  1.......     disabled
//				  0.......     enable
//				  .X......  INTEDG: Interrupt edge select
//				  .1......     riging edge
//				  .0......     failling edge
//				  ..X.....  T0CS: TMR0 clock source select
//				  ..1.....     Transition on GP2/T0CKI pin
//				  ..0.....     Internal instruction cycle
//				  ...X....  T0SE: TMR0 source edge select
//				  ...1....     high-to-low edge
//				  ...0....     low-to-high edge
//				  ....X...  PSA: Prescaler assignment
//				  ....1...     WDT
//				  ....0...     Timer0
//				  .....XXX  PS:  Prescaler assignment
//				  .....000     TMR0 rate 1:2 / WDT rate 1:1
//				  .....001               1:4 /          1:2
//				  .....010               1:8 /          1:4
//				  .....011              1:16 /          1:8
//				  .....100              1:32 /         1:16
//				  .....101              1:64 /         1:32
//				  .....110             1:128 /         1:64
//				  .....111             1:256 /        1:128
	OPTION =    0b11010100;		// Timer0=CLKO(4MHz/4)/32 -> 31.25KHz

//		*** Watchdock timer ***
//
//				  WDTCON - Watchdog control register
//				  ---01000  Initial
//				  ...XXXX.  WDTPS: Watchdog timer period select bits
//				  ...0000.     1:32
//				  ...0001.     1:64
//				  ...0010.     1:128
//				  ...0011.     1:256
//				  ...0100.     1:512
//				  ...0101.     1:1024
//				  ...0110.     1:2048
//				  ...0111.     1:4096
//				  ...1000.     1:8192
//				  ...1001.     1:16384
//				  ...1010.     1:32768
//				  ...1011.     1:65536
//				  .......X  WDTPS: Watchdog timer period select bits
//				  .......1     WDT is turned on
//				  .......0     WDT is turned off
	WDTCON =    0b00010101;		// 32.25KHz / 32.768KHz ~=~ 1sec 

//		*** GPIO port ***
//
//				  PORTA - General purpose I/O register A
//                xxx00000  Initial
//				  X.......  RA7: OSC1/CLKI
//				  .X......  RA6: OSC2/CLKO
//				  ..X.....  RA5: MCLR/Vpp
//				  ...X....  RA4: AN4/T0CKI
//				  ....X...  RA3: AN3/Vref+
//				  .....X..  RA2: AN2/Vref-
//				  ......X.  RA1: AN1
//				  .......X  RA0: AN0
	PORTA =     0b00000000;		// Initial all off

//				  TRISA - GPIO tri-state register A
//                11111111  Initial
//				  1.......  TRISA7: RA7 1=input only
//				  .0......  TRISA6: RA6 0=output only
//				  ..1.....  TRISA5: RA5 1=input only
//				  ...X....  TRISA4: RA4 1=input 0=output
//				  ....X...  TRISA3: RA3 1=input 0=output
//				  .....X..  TRISA2: RA2 1=input 0=output
//				  ......X.  TRISA1: RA1 1=input 0=output
//				  .......X  TRISA0: RA0 1=input 0=output
	TRISA =     0b10100011;		// RA2,3,4,6=output

//				  PORTB - General purpose I/O register B
//                xxxxxxxx  Initial
//				  X.......  RB7: AN6/T1OSI/PGD
//				  .X......  RB6: AN5/T1OSO/T1CKI/PGC
//				  ..X.....  RB5: SS/TX/CK
//				  ...X....  RB4: SCK/SCL
//				  ....X...  RB3: CCP1/PGM
//				  .....X..  RB2: SDO/RX/DT
//				  ......X.  RB1: SDI/SDA
//				  .......X  RB0: INT/CCP
	PORTB =     0b00000000;		// Initial all off

//				  TRISB - GPIO tri-state register B
//                11111111  Initial
//				  X.......  TRISB7: RB7 1=input 0=output
//				  .X......  TRISB6: RB6 1=input 0=output
//				  ..X.....  TRISB5: RB5 1=input 0=output
//				  ...X....  TRISB4: RB4 1=input 0=output
//				  ....X...  TRISB3: RB3 1=input 0=output
//				  .....X..  TRISB2: RB2 1=input 0=output
//				  ......X.  TRISB1: RB1 1=input 0=output
//				  .......X  TRISB0: RB0 1=input 0=output
	TRISB =     0b00000000;		// all output

//		*** Timer 0 ***
//
//				  TMR0 - Timer0 module register
//				  ????????  Initial
	TMR0 = 		0;				// 32.25KHz / 8bit = max122Hz

//		*** Timer 1 ***
//

//		*** Timer 2 ***
//

//		*** A/D Converter ***
//
//				  ADCON0 - A/D control register 0
//				  00000000  Initial
//				  XX......  ADCS: A/D converter clock select
//				  00......    ADCS2=0 Fosc/2   ADCS2=1 Fosc/4
//				  01......            Fosc/8           Fosc/16
//				  10......            Fosc/32          Fosc/64
//				  11......            FRC              FRC
//				  ..XXX...  CHS: Analog channel select bits
//				  ..000...    Channel 0 RA0/AN0
//				  ..001...    Channel 1 RA1/AN1
//				  ..010...    Channel 2 RA2/AN2
//				  ..011...    Channel 3 RA3/AN3
//				  ..100...    Channel 4 RA4/AN4
//				  ..101...    Channel 5 RB6/AN5
//				  ..110...    Channel 6 RB7/AN6
//				  .....X..  GO/DONE: A/D conversion status bit
//				  .....1..    ADON=1 in progress
//				  .....0..           not in progress
//				  .......X  ADON: A/D on bit
//				  .......1    module is operating
//				  .......0    module is shutoff
	ADCON0 =    0b11000000;		// FRC AN0 -> <Dynamic change>

//				  ADCON1 - A/D control register 1
//				  00000000  Initial
//				  X.......  ADFM: A/D result format select bit
//				  1.......    Right justified
//				  0.......    Left justified
//				  .X......  ADCS2: A/D clock divied by 2 select bit
//				  .1......    Enabled
//				  .0......    Disabled
//				  ..XX....  VCFG: A/D port configuration control bits
//				  ..00....    VREF+=AVdd  VREF-=AVss
//				  ..01....    VREF+=AVdd  VREF-=VREF-
//				  ..10....    VREF+=VREF+ VREF-=AVss
//				  ..11....    VREF+=VREF+ VREF-=VREF-
	ADCON1 =    0b10000000;		// Right adj, clock/1, Vref=Vdd:Vss

//				  ANSEL - Analog select register
//				  01111111  Initial
//				  .X......  ANS6 1=Analog 0=Digital
//				  ..X.....  ANS5 1=Analog 0=Digital
//				  ...X....  ANS4 1=Analog 0=Digital
//				  ....X...  ANS3 1=Analog 0=Digital
//				  .....X..  ANS2 1=Analog 0=Digital
//				  ......X.  ANS1 1=Analog 0=Digital
//				  .......X  ANS0 1=Analog 0=Digital
	ANSEL =     0b00000001;		// 

//		*** Interrupt ***
//
//				  PIE1 - Peripheral interrupt enable register 1
//                -0000000  Initial
//							0=disable 1=enable
//				  .X......  ADIE  : A/D Converter interrupt
//				  ..X.....  RCIE  : AUSART Receive interrupt
//				  ...X....  TXIE  : AUSART Transmit interrupt
//				  ....X...  SSPIE : SSP interrupt
//				  .....X..  CCP1IE: CCP1 interrupt
//				  ......X.  TMR2IE: Timer2 to PR2 match overflow
//				  .......X  TMR1IE: Timer1 overflow
	PIE1 =      0b00000000;		// Off

//				  PIE2 - Peripheral interrupt enable register 2
//                -0000000  Initial
//							0=disable 1=enable
//				  X.......  OSFIE : Oscillator fail
//				  .X......  CMIE  : Comparator interrupt
//				  ...X....  EEIE  : EEPROM Write complete
	PIE2 =      0b00000000;		// Off

//				  INTCON - Interrupt control register
//                 00000000  Initial
//							0=disable 1=enable
//				  X.......  GIE : Global interrupt
//				  .X......  PEIE: Peripheral intterupt -> PIE1
//				  ..X.....  T0IE: TMR0 Overflow interrupt
//				  ...X....  INTE: RB0/INT external interrupt
//				  ....X...  RBIE: RB Port Change interrupt
//							0=not(Appl reset) 1=occurred
//				  .....X..  T0IF: TMR0 Overflow interrupt flag
//				  ......X.  INTF: RB0/INT external interrupt flag
//				  .......X  GPIF: RB Port Change interrupt flag
	INTCON =    0b10100000;		// TMR0 enable

	//
	// Measure
	//
	psum = 0;
	isum = 0;
	endflg = 0;
	adc_start(0);
	for(sec = 0; sec < TIMELIM; sec++) {
		int vbat;						// Battery voltage
		int ibat;						// Battery current
		int pbat;						// Electricity consumption
		int vrs;						// Rsense voltage
		int a;
		int b;
		int i;

		a = 0;
		b = 0;
		for(i = 0; i < PLPF; i ++) {
			a += adc_read(0);
			b += adc_read(1);
			__delay_ms(10);
		}
		a = a / PLPF;
		b = b / PLPF;

		vbat = (long)a * VREF / 10 / 1024;	// [V*100] Res=10bit
		vrs  = (long)(a - b) * VREF / 1024;	// [mV]
		ibat = vrs;						// [mA] Rsense=1ohm
		pbat = vbat * ibat / 100;		// [mW]
		psum += pbat;					// [mWsec]
		isum += ibat;					// [mAsec]

		if (sec < 10) {					// Waiting
			ibat = 0;
		} else {						// Start
			if (vbat < VBATLOLIM) {
				if (endflg++ > 3) {		// A/D sampling error reject
					bitclr(PORTA, 6);	// Gate off
					break;
				}
			} else {
				bitset(PORTA, 6);		// Gate on
				endflg = 0;
			}
		}

		if ((unsigned char)sec & 0x01) {
			LED7SEG_DP(0,0);
			LED7SEG_DispInt(ibat, 1, 3);
		} else {
			LED7SEG_DispInt(vbat+1000, 1, 3);	// 1000:cancel zerosupress
			LED7SEG_DP(1,1);
		}

		while(upd == 0) {				// wait 1sec - Timer0 intr
			CLRWDT();
		}
		upd = 0;
	}
	adc_stop();
	LED7SEG_DP(0,0);

	//
	// Result
	//
	{
		int wh;
		int ah;
		int h;
		int i;

		wh = psum / 3600;					// [mWh]
		ah = isum / 3600;					// [mAh]
		h  = sec  / 3600;
		for(i = 0; ; i++) {
			int p;

			if (i & 0x2) {
				if (i & 0x1) {
					p = h;
				} else {
					p = ah;
				}
			} else {
				p = wh;
			}
			if (p <= 999) {
				LED7SEG_DispInt(p, 1, 3);
			} else if (p <= 9999) {
				LED7SEG_DispInt(p/10, 1, 3);
				LED7SEG_DP(1,1);
			} else {
				LED7SEG_DispInt(p/100, 1, 3);
				LED7SEG_DP(1,2);
			}
			__delay_ms(50);
			LED7SEG_PutPattern(0, 0);
			LED7SEG_DP(0,0);
			__delay_ms(10);
			#asm
			sleep
			#endasm
		}
	}
}